{
  TEmiDriveComboBox component

  Author  : Eric MINET - eric.minet@libertysurf.fr
            Web page : http://perso.libertysurf.fr/eminet/

  Version : 1.00

  Status  : Freeware with sources

  Target  : Delphi 4.0 should work on other 32 bits Delphi and C++ Builder versions

  Derived from TCustomComboBox

  This is a DriveComboBox with added features :

    - Allow all drive letters to be displayed
    - Filters for the drive types displayed (Options property)
    - Small or large icon display (IconSize property)
    - Extended informations on the selected drive
      ( SelectedDriveType, SelectedVolumeID and SelectedNetworkVolume properties)

  Deleted feature : no more direct link to TDirectoryListBox

  Future enhancements : detection of the floppy type and
    network volumes not available.

  The demo project into demo.zip allows you to manipulate all these features.
}

unit emiDriveCombo;

{$R-}

interface

uses Windows, Messages, SysUtils, Classes, Controls, Graphics, Forms,
  Menus, StdCtrls, Buttons;

{Component palette tab name for register method}
const ThePalettePage = 'Add-Ons';

type

{ TEmiDriveComboBox }

  TDriveType = (dtUnknown, dtNoDrive, dtFloppy, dtFixed, dtNetwork, dtCDROM,
    dtRAM);
  TDriveOption = (doNoFloppy, doNoFixed, doNoCDROM, doNoNetwork,
                  doNoRAM, doNoFree, doNoTape, doNoABDrives);
  TDriveOptions = set of TDriveOption;
  TTextCase = (tcLowerCase, tcUpperCase);
  TIconSize = (tiSmallIcon, tiLargeIcon);

  // Conteneur d'informations pour un disque de la liste
  TDriveInfos = class(TObject)
  private
    FDriveType     : TDriveType;
    FVolumeID      : string;
    FNetworkVolume : string;
  public
    constructor Create; virtual;
    property DriveType : TDriveType read FDriveType write FDriveType;
    property NetworkVolume : string read FNetworkVolume write FNetworkVolume;
    property VolumeID : string read FVolumeID write FVolumeID;
  end;

  TEmiDriveComboBox = class(TCustomComboBox)
  private
    FDrive      : Char;
    FIconSize   : TIconSize;
    // Liste de TDriveInfos synchronise avec la liste des disques
    FDriveInfos : TList;
    FOptions    : TDriveOptions;
    FTextCase   : TTextCase;
    procedure   CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    function    GetSelectedDriveType: TDriveType;
    function    GetSelectedVolumeID: string;
    function    GetSelectedNetworkVolume:string;
    procedure   SetDrive(NewDrive: Char);
    procedure   SetIconSize(NewIconSize: TIconSize);
    procedure   SetOptions(Value: TDriveOptions);
    procedure   SetTextCase(NewTextCase: TTextCase);
    procedure   ReadBitmaps;
    procedure   FreeBitmaps;
    procedure   FreeDriveInfos;
    procedure   ResetItemHeight;
  protected
    LFloppyBMP, SFloppyBMP, FixedBMP,
    NetworkBMP, CDROMBMP, RAMBMP,
    TapeBMP, NoDriveBMP, NoNetBMP : TBitmap;
    procedure   CreateWnd; override;
    procedure   DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure   Click; override;
    procedure   BuildList; virtual;
    procedure   Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    property    Text;
    property    Drive: Char read FDrive write SetDrive;
    property    SelectedDriveType: TDriveType read GetSelectedDriveType;
    property    SelectedVolumeID: string read GetSelectedVolumeID;
    property    SelectedNetworkVolume:string read GetSelectedNetworkVolume;
  published
    property    Color;
    property    Ctl3D;
    property    DragMode;
    property    DragCursor;
    property    Enabled;
    property    Font;
    property    IconSize: TIconSize read FIconSize write SetIconSize default tiSmallIcon;
    property    ImeMode;
    property    ImeName;
    property    Options: TDriveOptions read FOptions write SetOptions default [];
    property    ParentColor;
    property    ParentCtl3D;
    property    ParentFont;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    TextCase: TTextCase read FTextCase write SetTextCase default tcLowerCase;
    property    Visible;
    property    OnChange;
    property    OnClick;
    property    OnDblClick;
    property    OnDragDrop;
    property    OnDragOver;
    property    OnDropDown;
    property    OnEndDrag;
    property    OnEnter;
    property    OnExit;
    property    OnKeyDown;
    property    OnKeyPress;
    property    OnKeyUp;
    property    OnStartDrag;
  end;

procedure Register;

implementation

{$R emiDriveCombo}

//////////////////////////////////////////////////////////////////////
//       Conteneur d'informations pour un disque de la liste        //
//////////////////////////////////////////////////////////////////////
constructor TDriveInfos.Create;
begin
    FDriveType     := dtUnknown;
    FVolumeID      := '';
    FNetworkVolume := '';
end;

//////////////////////////////////////////////////////////////////////
//             TEmDriveComboBox - mthodes gnrales                //
//////////////////////////////////////////////////////////////////////
function VolumeID(DriveChar: Char): string;
var
  OldErrorMode: Integer;
  NotUsed, VolFlags: DWORD;
  Buf: array [0..MAX_PATH] of Char;
begin
  OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
  try
    Buf[0] := #$00;
    if GetVolumeInformation(PChar(DriveChar + ':\'), Buf, sizeof(Buf),
      nil, NotUsed, VolFlags, nil, 0) then
      SetString(Result, Buf, StrLen(Buf))
    else Result := '';
    if DriveChar < 'a' then
      Result := AnsiUpperCaseFileName(Result)
    else
      Result := AnsiLowerCaseFileName(Result);
    Result := Format('[%s]',[Result]);
  finally
    SetErrorMode(OldErrorMode);
  end;
end;

function NetworkVolume(DriveChar: Char): string;
var
  Buf: Array [0..MAX_PATH] of Char;
  DriveStr: array [0..3] of Char;
  BufferSize: DWORD;
begin
  BufferSize := sizeof(Buf);
  DriveStr[0] := UpCase(DriveChar);
  DriveStr[1] := ':';
  DriveStr[2] := #0;
  if WNetGetConnection(DriveStr, Buf, BufferSize) = WN_SUCCESS then
  begin
    SetString(Result, Buf, BufferSize);
    if DriveChar < 'a' then
      Result := AnsiUpperCaseFileName(Result)
    else
      Result := AnsiLowerCaseFileName(Result);
  end
  else
    Result := VolumeID(DriveChar);
end;

function GetItemHeight(Font: TFont): Integer;
var
  DC: HDC;
  SaveFont: HFont;
  Metrics: TTextMetric;
begin
  DC := GetDC(0);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  Result := Metrics.tmHeight;
end;

//////////////////////////////////////////////////////////////////////
//              TEmiDriveComboBox - implmentation                  //
//////////////////////////////////////////////////////////////////////
constructor TEmiDriveComboBox.Create(AOwner: TComponent);
var
  Temp: ShortString;
begin
  inherited Create(AOwner);
  FDriveInfos:=TList.Create;
  IconSize:=tiSmallIcon;
  FOptions:=[];
  Style := csOwnerDrawFixed;
  ReadBitmaps;
  GetDir(0, Temp);
  FDrive := Temp[1]; { make default drive selected }
  if FDrive = '\' then FDrive := #0;
  ResetItemHeight;
end;

destructor TEmiDriveComboBox.Destroy;
begin
  FreeBitmaps;
  FreeDriveInfos;
  FDriveInfos.Free;
  inherited Destroy;
end;

procedure TEmiDriveComboBox.Loaded;
begin
  inherited Loaded;
  ResetItemHeight;
end;

procedure TEmiDriveComboBox.SetOptions(Value: TDriveOptions);
var OldDrive: Char;
begin
  FOptions:=Value;
  if not(csDesigning in ComponentState) then
  begin
    OldDrive := FDrive;
    BuildList;
    SetDrive (OldDrive);
  end;
end;

procedure TEmiDriveComboBox.BuildList;
var
  DriveNum   : Integer;
  DriveStart : integer;
  DriveChar  : Char;
  DriveType  : TDriveType;

  procedure AddDrive(ADrive : Char; const VolName, NetName: string;
    ABitmap: TObject; ADriveType : TDriveType);
  var DriveInfo  : TDriveInfos;
      s          : string;
  begin
    DriveInfo:=TDriveInfos.Create;
    try
      DriveInfo.DriveType:=ADriveType;
      DriveInfo.VolumeID:=VolName;
      DriveInfo.NetworkVolume:=NetName;
      if VolName = ''
      then
        begin
          if NetName = ''
          then s:=Format('%s:',[DriveChar])
          else s:=Format('%s: %s',[DriveChar, NetName])
        end
      else s:=Format('%s: %s',[DriveChar, VolName]);
      Items.AddObject(s, ABitmap);
      FDriveInfos.Add(DriveInfo);
    except
      DriveInfo.Free;
    end;
  end;

begin
  // Effacer la liste de la combo et les informations associes
  Clear;
  FreeDriveInfos;
  // Ajustement si lecteurs A et B ignors
  if (doNoABDrives in FOptions) then DriveStart:=2 else DriveStart:=0;
  // Boucle sur les disques
  for DriveNum := DriveStart to 25 do
  begin
    // Lettre du disque
    DriveChar := Char(DriveNum + Ord('a'));
    // type de disque
    DriveType := TDriveType(GetDriveType(PChar(DriveChar + ':\')));
    // Mise en majuscules si ncessaire
    if TextCase = tcUpperCase then DriveChar := Upcase(DriveChar);
    // rpartition
    case DriveType of
      dtUnknown :  if not(doNoTape in FOptions)
                   then AddDrive(DriveChar, '', '', TapeBMP, DriveType);
                   //Items.AddObject(DriveChar + ':', TapeBMP);
      dtNoDrive :  if not(doNoFree in FOptions)
                   then AddDrive(DriveChar, '', '', NoDriveBMP, DriveType);
                   //Items.AddObject(DriveChar + ':', NoDriveBMP);
      dtFloppy  :  if not(doNoFloppy in FOptions)
                   then AddDrive(DriveChar, '', '', SFloppyBMP, DriveType);
                   //Items.AddObject(DriveChar + ':', SFloppyBMP);
      dtFixed   :  if not(doNoFixed in FOptions)
                   then AddDrive(DriveChar, VolumeID(DriveChar), '', FixedBMP, DriveType);
                   //AddDrive(VolumeID(DriveChar), FixedBMP);
      dtNetwork :  if not(doNoNetwork in FOptions)
                   then AddDrive(DriveChar, '', NetworkVolume(DriveChar), NetworkBMP, DriveType);
                   //AddDrive(NetworkVolume(DriveChar), NetworkBMP);
      dtCDROM   :  if not(doNoCDROM in FOptions)
                   then AddDrive(DriveChar, VolumeID(DriveChar), '', CDROMBMP, DriveType);
                   //AddDrive(VolumeID(DriveChar), CDROMBMP);
      dtRAM     :  if not(doNoRAM in FOptions)
                   then AddDrive(DriveChar, VolumeID(DriveChar), '', RAMBMP, DriveType);
                   //AddDrive(VolumeID(DriveChar), RAMBMP);
    end;
  end;
end;

procedure TEmiDriveComboBox.SetDrive(NewDrive: Char);
var
  Item: Integer;
  drv: string;
begin
  if (ItemIndex < 0) or (UpCase(NewDrive) <> UpCase(FDrive)) then
  begin
    if NewDrive = #0
    then
      begin
        FDrive := NewDrive;
        ItemIndex := -1;
      end
    else
      begin
        if TextCase = tcUpperCase
        then FDrive := UpCase(NewDrive)
        else FDrive := Chr(ord(UpCase(NewDrive)) + 32);
        { change selected item }
        for Item := 0 to Items.Count - 1 do
        begin
          drv := Items[Item];
          if (UpCase(drv[1]) = UpCase(FDrive)) and (drv[2] = ':') then
          begin
            ItemIndex := Item;
            break;
          end;
        end;
        if (ItemIndex<0) and (Items.Count>0) then ItemIndex:=0;
      end;
    Change;
  end;
end;

procedure TEmiDriveComboBox.SetIconSize(NewIconSize: TIconSize);
var
  OldDrive: Char;
begin
  if FIconSize<>NewIconSize then
  begin
    FIconSize := NewIconSize;
    OldDrive := FDrive;
    FreeBitmaps;
    ReadBitmaps;
    ResetItemHeight;
    BuildList;
    SetDrive (OldDrive);
    RecreateWnd;
  end;
end;

procedure TEmiDriveComboBox.SetTextCase(NewTextCase: TTextCase);
var
  OldDrive: Char;
begin
  if FTextCase<>NewTextCase then
  begin
    FTextCase := NewTextCase;
    OldDrive := FDrive;
    BuildList;
    SetDrive (OldDrive);
  end;
end;

procedure TEmiDriveComboBox.CreateWnd;
begin
  inherited CreateWnd;
  BuildList;
  SetDrive (FDrive);
end;

procedure TEmiDriveComboBox.DrawItem(Index: Integer; Rect: TRect;
  State: TOwnerDrawState);
var
  Bitmap: TBitmap;
  bmpWidth: Integer;
begin
  with Canvas do
  begin
    FillRect(Rect);
    bmpWidth  := 16;
    Bitmap := TBitmap(Items.Objects[Index]);
    if Bitmap <> nil then
    begin
      bmpWidth := Bitmap.Width;
      BrushCopy(Bounds(Rect.Left + 2,
               (Rect.Top + Rect.Bottom - Bitmap.Height) div 2,
               Bitmap.Width, Bitmap.Height),
               Bitmap, Bounds(0, 0, Bitmap.Width, Bitmap.Height),
               Bitmap.Canvas.Pixels[0, Bitmap.Height - 1]);
    end;
     { uses DrawText instead of TextOut in order to get clipping against
       the combo box button   }
    Rect.Left := Rect.Left + bmpWidth + 6;
    DrawText(Canvas.Handle, PChar(Items[Index]), -1, Rect,
             DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
  end;
end;

procedure TEmiDriveComboBox.Click;
begin
  inherited Click;
  if ItemIndex >= 0 then Drive := Items[ItemIndex][1];
end;

procedure TEmiDriveComboBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  ResetItemHeight;
  RecreateWnd;
end;

procedure TEmiDriveComboBox.ResetItemHeight;
var
  nuHeight: Integer;
begin
  nuHeight :=  GetItemHeight(Font);
  if nuHeight < (LFloppyBMP.Height) then nuHeight := LFloppyBmp.Height;
  ItemHeight := nuHeight;
end;

procedure TEmiDriveComboBox.FreeBitmaps;
begin
  LFloppyBMP.Free;
  SFloppyBMP.Free;
  FixedBMP.Free;
  TapeBMP.Free;
  NetworkBMP.Free;
  NoNetBMP.Free;
  CDROMBMP.Free;
  RAMBMP.Free;
  NoDriveBMP.Free;
end;

procedure TEmiDriveComboBox.FreeDriveInfos;
var i : integer;
begin
     for i:=0 to FDriveInfos.Count-1 do
       TDriveInfos(FDriveInfos.Items[i]).Free;
     FDriveInfos.Clear;
end;

function TEmiDriveComboBox.GetSelectedDriveType: TDriveType;
begin
     result:=dtUnknown;
     if (ItemIndex<>-1) and (FDriveInfos.Count>ItemIndex)
     then result:=TDriveInfos(FDriveInfos.Items[ItemIndex]).DriveType;
end;

function TEmiDriveComboBox.GetSelectedVolumeID: string;
begin
     result:='';
     if (ItemIndex<>-1) and (FDriveInfos.Count>ItemIndex)
     then result:=TDriveInfos(FDriveInfos.Items[ItemIndex]).VolumeID;
end;

function TEmiDriveComboBox.GetSelectedNetworkVolume:string;
begin
     result:='';
     if (ItemIndex<>-1) and (FDriveInfos.Count>ItemIndex)
     then result:=TDriveInfos(FDriveInfos.Items[ItemIndex]).NetworkVolume;
end;

procedure TEmiDriveComboBox.ReadBitmaps;
const ResNames : array[0..17] of string =
      ('LARGEFLOPPY16', 'SMALLFLOPPY16', 'HARDDRIVE16', 'NETCONNECT16',
       'NETUNCONNECT16', 'CDROM16', 'RAMDRIVE16', 'TAPEDRIVE16', 'NODRIVE16',
       'LARGEFLOPPY32', 'SMALLFLOPPY32', 'HARDDRIVE32', 'NETCONNECT32',
       'NETUNCONNECT32', 'CDROM32', 'RAMDRIVE32', 'TAPEDRIVE32', 'NODRIVE32');
var s    : string;
    Offs : integer;
begin
  if FIconSize=tiSmallIcon then Offs:=0 else Offs:=9;
  { assign bitmap glyphs }
  s:=ResNames[Offs];
  LFloppyBMP := TBitmap.Create;
  LFloppyBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+1];
  SFloppyBMP := TBitmap.Create;
  SFloppyBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+2];
  FixedBMP := TBitmap.Create;
  FixedBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+3];
  NetworkBMP := TBitmap.Create;
  NetworkBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+4];
  NoNetBMP := TBitmap.Create;
  NoNetBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+5];
  CDROMBMP := TBitmap.Create;
  CDROMBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+6];
  RAMBMP := TBitmap.Create;
  RAMBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+7];
  TapeBMP := TBitmap.Create;
  TapeBMP.Handle := LoadBitmap(HInstance, PChar(s));
  s:=ResNames[Offs+8];
  NoDriveBMP := TBitmap.Create;
  NoDriveBMP.Handle := LoadBitmap(HInstance, PChar(s));
end;

procedure Register;
begin
  RegisterComponents(ThePalettePage, [TEmiDriveComboBox]);
end;

end.

